// $Id: CheckModule.java,v 1.14 2004/06/14 21:02:32 m31 Exp $
//
// This file is part of the program suite "Principia Mathematica II"
// which is a working prototype for the main project:
// "Hilbert II" - http://www.qedeq.org
//
// Copyright 2001-2004  Michael Meyling <mime@qedeq.org>.
//
// "Principia Mathematica II" is free software; you can redistribute
// it and/or modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.

package com.meyling.principia;


import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import com.meyling.principia.io.ParsingException;
import com.meyling.principia.io.Utility;
import com.meyling.principia.module.ModuleContext;


/**
 * This class checks Principia Mathematica II qedeq modules for errors.
 *
 * @version $Revision: 1.14 $
 * @author  Michael Meyling
 */
public class CheckModule {


    /**
     * Writes calling convention to <code>System.err</code>.
     */
    private static void info() {
        System.err.println("Name");
        System.err.println("----");
        System.err.println(Utility.getClassName(CheckModule.class) + " - proof check URLs");
        System.err.println();
        System.err.println("Synopis");
        System.err.println("-------------------");
        System.err.println(CheckModule.class.getName() + " [-h] [<module_address>]...");
        System.err.println();
        System.err.println("Description");
        System.err.println("-----------");
        System.err.println("This program checks *Principia Mathematica II* modules for errors. Such a module");
        System.err.println("contains mathematical knowledge in a checkable form. This programm tries to");
        System.err.println("load the modules by the given addresses. All neccassary other modules");
        System.err.println("(\"IMPORTS\") are also loaded. If any loaded module doesn't pass the");
        System.err.println("verification check, the result is negative and the position and reason of the");
        System.err.println("problem is given (path to local copy \":\" line number \":\" column number).");
        System.err.println("Any new module is first copied into a local file buffer and then loaded");
        System.err.println("into memory. Further attempts to load the module lead to a search in the memory");
        System.err.println("and next in the local file buffer.");
        System.err.println("If you want to create a fresh copy, you have to delete the file buffer manually");
        System.err.println("(Normal file location: \"" + ModuleContext.getBufferDirectory() + "\").");
        System.err.println();
        System.err.println("Options");
        System.err.println("-------");
        System.err.println("-h\twrites this text");
        System.err.println("\taddress of a module to check, e.g.:");
        System.err.println("\thttp://www.qedeq.org/" + ModuleContext.getVersionDirectory() + "/proptheo1_1.00.00_1.00.00.qedeq");
        System.err.println();
        System.err.println("Examples");
        System.err.println("--------");
        System.err.println(CheckModule.class.getName() + " http://www.qedeq.org/" + ModuleContext.getVersionDirectory() + "/proptheo1_1.00.00_1.00.00.qedeq");
        System.err.println();
        System.err.println("Prototype Warning");
        System.err.println("-----------------");
        System.err.println("This program suite is only a working prototype for the main project");
        System.err.println("*Hilbert II*. It concentrates on the basic logic foundation of");
        System.err.println("(mathematics. For more information about *Hilbert II* look at:");
        System.err.println("http://www.qedeq.org/");
    }


    /**
     * Check modules by trying to load them.
     *
     * @param   args addresses of modules and options
     */
    public static void main(final String[] args) {

        try {
            PmiiFacade.init();
            doIt(args);
            PmiiFacade.shutdown();
        } catch (Exception e) {
            e.printStackTrace();
        }
    }


    /**
     * Check modules by trying to load them.
     *
     * @param   args addresses of modules and options
     */
    public static void doIt(final String[] args) {

        final List address = new ArrayList();

        if (args.length == 0) {
            info();
            return;
        }
        for (int i = 0; i < args.length; i++) {
            if (args[i].startsWith("-")) {  // option
                final String option = args[i].substring(1).toLowerCase();
                if (option.equals("help") || option.equals("h")
                        || option.equals("?")) {
                    info();
                } else {                    // unknown option
                    System.err.println("\nUnknown option: " + option);
                    info();
                    return;
                }
            } else {                        // no option, must be file name
                address.add(args[i]);
            }
        }
        for (int i = 0; i < address.size(); i++) {
            checkModule((String) address.get(i));
        }
    }


    /**
     * Check a module.
     *
     * @param   address  address of module
     * @return  is module ok?
     */
     public final static boolean checkModule(final String address) {
        try {
            PmiiFacade.getInstance().getModuleContext().loadModule(address);
            System.out.println("checked:");
            System.out.println("  " + address);
            return true;
        } catch (IOException e) {
            System.err.println(e.getMessage());
        } catch (IllegalArgumentException e) {
            e.printStackTrace(System.err); 
            System.err.println(e.getMessage());
        } catch (ParsingException e) {
            System.err.println(new File(e.getAddress())
                .getAbsolutePath() + ":" + e.getRow() + ":"
                + e.getColumn() + ":");
            System.err.println(e.getDescription());
            System.err.println(e.getLine());
            final StringBuffer pointer = Utility.getSpaces(e.getColumn() - 1);
            pointer.append("^");
            System.err.println(pointer);
        }
        return false;
    }


}
