// $Id: ReplacePredicateVariable.java,v 1.9 2004/02/27 16:29:56 m31 Exp $
//
// This file is part of the program suite "Principia Mathematica II"
// which is a working prototype for the main project:
// "Hilbert II" - http://www.qedeq.org
//
// Copyright 2001-2004  Michael Meyling <mime@qedeq.org>.
//
// "Principia Mathematica II" is free software; you can redistribute
// it and/or modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.

package com.meyling.principia.logic.rule;


import com.meyling.principia.argument.AbstractArgumentList;
import com.meyling.principia.argument.Argument;
import com.meyling.principia.argument.ArgumentException;
import com.meyling.principia.argument.Counter;
import com.meyling.principia.argument.Text;
import com.meyling.principia.logic.basic.BasicPatternVariables;
import com.meyling.principia.logic.basic.BasicSubjectVariablePatternVariable;
import com.meyling.principia.logic.basic.Formula;
import com.meyling.principia.logic.basic.PredicateVariable;
import com.meyling.principia.logic.basic.RegularArgumentList;
import com.meyling.principia.logic.paragraph.ParagraphCreator;
import com.meyling.principia.logic.paragraph.ProofLineList;
import com.meyling.principia.logic.paragraph.RuleDeclaration;
import com.meyling.principia.module.Module;
import com.meyling.principia.module.Version;


/**
 * Replace predicate variable by a formula:
 * <p></p>
 * <p><code>       A[R(x,y,z)]</code></p>
 * <p><code>  formula F(x,y,z)</code></p>
 * <p><code>------------------</code></p>
 * <p><code>       A[F(x,y,z)]</code></p>
 *
 * @version $Revision: 1.9 $
 * @author  Michael Meyling
 */
public class ReplacePredicateVariable extends AbstractArgumentList
        implements Argument, Rule {

    /** minimum version required for using this rule */
    private final static Version version;

    /** proof line number to operate on */
    private final int n;

    /** predicate variable which will be replaced */
    private final PredicateVariable prop;

    /** formula to replace the predicate variable with */
    private final Formula replacement;


    /**
     * Initialize version information.
     */
    static {
        try {
            version = new Version(
                    new Argument[] {new Text("1.00.00")});
        } catch (ArgumentException e) {
            throw new IllegalArgumentException(e.toString());
        }
    }


    /**
     * Constructs a replace predicate variable rule.
     *
     * @param   arguments   <ul><li>0: Counter as an proof line
     *                      reference</li>
     *                      <li>1: predicate variable, that contains only
     *                      pattern variables as arguments. All of these
     *                      pattern variables must be different, so that
     *                      the number of different pattern variables in
     *                      the predicate variable formula is equal to
     *                      the list length of the predicate variable.</li>
     *                      <li>2: formula that must have the same
     *                      pattern variables</li>
     *
     * @throws  ArgumentException    if there are not three instances of
     *                      {@link Counter}, {@link PredicateVariable},
     *                      {@link Formula} (in exactly that order) or
     *                      the restrictions from above are not fullfilled
     *
     */
    public ReplacePredicateVariable(final Argument[] arguments)
            throws ArgumentException {
        super (arguments);
        if (arguments.length != 3) {
            throw new ArgumentException(
                RuleConstants.CODE_CONSTRUCTOR_WRONG_ARGUMENT_NUMBER,
                RuleCreator.getName(this.getClass())
                + RuleConstants.OPERATOR_WITH_THREE_ARGUMENTS);
        }
        if (!(arguments[0] instanceof Counter)) {
            throw new ArgumentException(
                RuleConstants.CODE_CONSTRUCTOR_WRONG_ARGUMENT_TYPE,
                RuleCreator.getName(this.getClass())
                + RuleConstants.FIRST_ARGUMENT_TYPE
                + RuleConstants.NUMBER,
                arguments[0]);
        }
        this.n = ((Counter) arguments[0]).getNumber() - 1;
        if (n < 0) {
            throw new ArgumentException(
                RuleConstants.CODE_CONSTRUCTOR_WRONG_ARGUMENT_TYPE,
                RuleCreator.getName(this.getClass())
                + RuleConstants.FIRST_ARGUMENT_TYPE
                + RuleConstants.POSITIVE_INTEGER_EXPECTED,
                arguments[0]);
        }
        if (!(arguments[1] instanceof PredicateVariable)) {
            throw new ArgumentException(
                RuleConstants.CODE_CONSTRUCTOR_WRONG_ARGUMENT_TYPE,
                RuleCreator.getName(this.getClass())
                + RuleConstants.SECOND_ARGUMENT_TYPE
                + RuleCreator.getName(PredicateVariable.class),
                arguments[1]);
        }
        this.prop = (PredicateVariable) arguments[1];
        // check if predicate variable has a number
        if (!(prop.getArgument(0) instanceof Counter)) {
            throw new ArgumentException(
                RuleConstants.CODE_CONSTRUCTOR_WRONG_ARGUMENT_TYPE,
                RuleCreator.getName(this.getClass())
                + RuleConstants.SECOND_ARGUMENT
                + RuleConstants.MUST_HAVE_COUNTER,
                prop.getArgument(0));
        }
        final RegularArgumentList list
            = (RegularArgumentList) prop.getArgument(1);
        for (int i = 0; i < list.getArgumentSize(); i++) {
            if (!(list.getArgument(i) instanceof
                    BasicSubjectVariablePatternVariable)) {
                throw new ArgumentException(
                    RuleConstants.CODE_CONSTRUCTOR_WRONG_ARGUMENT_TYPE,
                    RuleCreator.getName(this.getClass())
                    + RuleConstants.SECOND_ARGUMENT
                    + " " + RuleCreator.getName(
                    BasicSubjectVariablePatternVariable.class),
                    list.getArgument(i));
            }
        }
        final BasicPatternVariables variables
            = new BasicPatternVariables(prop.getPatternVariables());
        // check if all pattern variables are different
        if (list.getArgumentSize()
                != variables.getSize()) {
            throw new ArgumentException(
                RuleConstants.CODE_CONSTRUCTOR_WRONG_ARGUMENT_TYPE,
                RuleCreator.getName(this.getClass())
                + RuleConstants.SECOND_ARGUMENT
                + RuleConstants.DIFFERENT_PATTERN_VARIABLES,
                prop);
        }
        if (!(arguments[2] instanceof Formula)) {
            throw new ArgumentException(
                RuleConstants.CODE_CONSTRUCTOR_WRONG_ARGUMENT_TYPE,
                RuleCreator.getName(this.getClass())
                + RuleConstants.THIRD_ARGUMENT_TYPE
                + RuleConstants.FORMULA,
                arguments[2]);
        }
        replacement = (Formula) arguments[2];
        // check if pattern variables in formula are the same
        if (!replacement.getPatternVariables().equals(variables)) {
            throw new ArgumentException(
                RuleConstants.CODE_CONSTRUCTOR_WRONG_ARGUMENT_TYPE,
                RuleCreator.getName(this.getClass())
                + RuleConstants.THIRD_ARGUMENT
                + RuleConstants.SAME_PATTERN_VARIABLES,
                arguments[2]);
        }

// TODO no more neccessary ???
/*
        // check if all pattern variables stand for subject variables
        try {
            variables.fillWithSubjectVariables(
                replacement.getSubjectVariables().getHighestNumber());
        } catch (IllegalArgumentException e) {
            throw new IllegalArgumentException(
                RuleCreator.getName(this.getClass())
                + RuleConstants.THIRD_ARGUMENT
                + RuleConstants.PATTERN_VARIABLES_MUST_STAND_FOR_SUBJECT_VARIABLES);
        }
        try {
            replacement.replace(variables);
        } catch (ArgumentException e) {
            throw new IllegalArgumentException(e.getMessage());
        }
*/
    }


    /**
     * Check if proof could be extended with formula if replacing the
     * predicate variable by the formula of the constructor.
     *
     * @param   module   module
     * @param   proofLines  proof lines
     * @param   position number of proof lines that could be used
     *                    in the rule
     * @param   formula  formula to add
     * @throws  ArgumentException    if addition is not possible
     *          because the proof line number is not allowed (e.g.
     *          between 1 and the minimum of <code>position</code>
     *          and current maximum proof line number)
     *          or the referenced proof line is no formula
     *          or the set of free variables of the formula and
     *           the set of bound variables of the proof line
     *           aren't disjunct
     *          or the set of bound variables
     *           of the formula and the subject variables of the
     *           proof line aren't disjunct
     *          or the predicate variable is in reach of an
     *           quantor with subject variable x, and x
     *           occurs in the replacement formula
     *          or the resulting formula is not equal to
     *          <code>formula</code>
     *          or a version conflict occurs
     */
    public final void check(final Module module,
            final ProofLineList proofLines,
            final int position,
            final Formula formula) throws ArgumentException {

        if (version.greaterThan(module.getRuleVersion())) {
            throw new ArgumentException(
                RuleConstants.CODE_METHOD_ARGUMENT_CHECK_FAILED,
                RuleCreator.getName(this.getClass())
                    + " " + RuleConstants.RULE_VERSION_CONFLICT
                    + version.getText(),
                this);
        }
        if (n < 0 || n >= proofLines.getArgumentSize() || n >= position) {
            throw new ArgumentException(
                RuleConstants.CODE_METHOD_ARGUMENT_CHECK_FAILED,
                RuleCreator.getName(this.getClass())
                    + RuleConstants.PROOF_LINES_OUT_OF_RANGE,
                this.getArgument(0));
        }
        final Formula proofLine
            = (Formula) proofLines.getArgument(n).getArgument(0);

        if (!proofLine.getSubjectVariables().emptyIntersection(
                replacement.getBoundSubjectVariables())
                || !proofLine.getBoundSubjectVariables().emptyIntersection(
                replacement.getFreeSubjectVariables())) {
            throw new ArgumentException(
                RuleConstants.CODE_METHOD_ARGUMENT_CHECK_FAILED,
                RuleCreator.getName(this.getClass())
                    + RuleConstants
                    .ALL_BOUND_OR_BOUND_FREE_NOT_EMPTY_INTERSECTION,
                this,
                proofLines.getArgument(n).getArgument(0));
        }
        final Formula replaced;
        try {
            replaced = (Formula) proofLine.replaceMatches(prop,
                replacement);
            // the missing condition was checked by constructing the
            //  replacement

        } catch (ArgumentException e) {     // e.g. if variable
            throw new ArgumentException(    // is double bound
                RuleConstants.CODE_METHOD_ARGUMENT_CHECK_FAILED,
                RuleCreator.getName(this.getClass())
                + RuleConstants.REPLACING_PREDICATE_VARIABLE_NOT_POSSIBLE,
                this,
                proofLines.getArgument(n).getArgument(0));
        }
        if (!replaced.equals(formula)) {
            throw new ArgumentException(
                RuleConstants.CODE_METHOD_ARGUMENT_CHECK_FAILED,
                RuleCreator.getName(this.getClass())
                    + RuleConstants.RESULT_NOT_EQUAL_WITH_EXPECTED,
                this,
                proofLines.getArgument(n).getArgument(0));
        }
    }


    /** @inherited */
    public ProofLineList extendWithout(final Module module,
            final ProofLineList proofLines, final int position)
            throws UnsupportedOperationException {
        throw new UnsupportedOperationException (
            "this rule is neccessary: "
            + RuleCreator.getName(this.getClass()));
    }


    /** @inherited */
    public int[] getProofLines() {
        return new int[] {this.n};
    }


    /** @inherited */
    public Rule changeProofLines(final int[] mapping) {
        try {
            return (Rule) create(new Argument[] {
                new Counter(mapping[n] + 1), prop, replacement});
        } catch (ArgumentException e) {
            throw new IllegalArgumentException(e.toString());
        } catch (IndexOutOfBoundsException e) {
            throw new IllegalArgumentException(e.toString());
        }
    }


    /**
     * Check this rule could be declared.
     *
     * @param   module      the module context
     * @param   declaration the declaration that shall declare this rule
     * @throws  IllegalArgumentException    if an programming error occured
     * @throws  ArgumentException    if declaration failed
     */
    public final static void checkDeclaration(final Module module,
        final RuleDeclaration declaration) throws ArgumentException {
        if (declaration.getArgumentSize() > 2) {
            throw new ArgumentException(
                RuleConstants.CODE_METHOD_ARGUMENT_CHECK_FAILED,
                ParagraphCreator.getName(RuleDeclaration.class)
                    + RuleConstants.OPERATOR_WITH_TWO_ARGUMENTS,
                declaration);
        }
    }


    /** @inherited */
    public final Version getVersion() {
        return ReplacePredicateVariable.version;
    }


    /** @inherited */
    public final Argument create(final Argument[] arguments)
            throws ArgumentException {
        return new ReplacePredicateVariable(arguments);
    }


    /** @inherited */
    public final String toString() {
        return "ReplacePredicate " + getArgument(0) + getArgument(1)
            + getArgument(2);
    }


}
